"""Provides common base classes that allow modules to talk to each other."""
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum
from itertools import chain
from typing import List, Optional, Sequence, Type, TypeVar, Union

from pydantic import Field, ValidationError
from pydantic.class_validators import validator
from pydantic.main import BaseModel
from pydicom.datadict import tag_for_keyword
from pydicom.dataset import Dataset

from dicomtrolley.exceptions import (
    DICOMTrolleyError,
    UnSupportedParameterError,
)


class DICOMObjectLevels:
    """DICOM uses a study->series->instance structure"""

    STUDY = "Study"
    SERIES = "Series"
    INSTANCE = "Instance"

    all = [STUDY, SERIES, INSTANCE]


class DICOMDownloadable:
    """An object that can be downloaded by a Downloader"""

    def reference(self) -> "DICOMObjectReference":
        raise NotImplementedError


@dataclass
class DICOMObjectReference(DICOMDownloadable):
    """Points to a study, series, or instance by uid

    Contrary to DICOMObjects, DICOMObjectReferences are shallow. They do
    not have parents or children and do not contain any additional information.

    They were created to use in object-level downloads (download study X)
    """

    study_uid: str

    @property
    def level(self) -> str:
        """Does this reference point to Study, Series or Instance?
        Returns
        -------
        str
            One of the values in DICOMObjectLevels.all
        """
        return ""

    def reference(self):
        return self


@dataclass
class InstanceReference(DICOMObjectReference):
    """All information needed to download a single slice (SOPInstance)"""

    study_uid: str
    series_uid: str
    instance_uid: str

    def __str__(self):
        return (
            f"InstanceReference {self.study_uid} -> {self.series_uid} "
            f"-> {self.instance_uid}"
        )

    @property
    def level(self):
        return DICOMObjectLevels.INSTANCE


@dataclass
class SeriesReference(DICOMObjectReference):
    """Reference to a single Series, part of a study"""

    study_uid: str
    series_uid: str

    def __str__(self):
        return f"SeriesReference {self.study_uid} -> {self.series_uid} "

    @property
    def level(self):
        return DICOMObjectLevels.SERIES


@dataclass
class StudyReference(DICOMObjectReference):
    """Reference to a single study"""

    study_uid: str

    def __str__(self):
        return f"StudyReference {self.study_uid}"

    @property
    def level(self):
        return DICOMObjectLevels.STUDY


class DICOMObject(BaseModel, DICOMDownloadable):
    """An object in the DICOM world. Base for Study, Series, Instance.

    dicomtrolley search methods always return instances based on DICOMObject
    dicomtrolley download methods take instances based on DICOMObject as input

    DICOMObjects know where they are in the DICOM tree; what their parents and
    children are. They can also contain additional information in the form of
    a DICOM dataset
    """

    class Config:
        arbitrary_types_allowed = True  # allows the use of Dataset type below

    uid: str
    data: Dataset

    def __str__(self):
        return type(self).__name__ + " " + self.uid

    def children(self):
        """

        Returns
        -------
        Sequence[DICOMObject]
            All direct child objects or empty list if none. Useful for iterating
        """
        raise NotImplementedError()

    def root(self):
        """The parent or parents parent or etc.

        Returns
        -------
        DICOMObject
        """
        raise NotImplementedError()

    def all_instances(self) -> List["Instance"]:
        """

        Returns
        -------
        List[Instance]
            All instances contained in this object
        """
        raise NotImplementedError()

    def all_series(self) -> List["Series"]:
        """

        Returns
        -------
        List[Series]
            All Series contained in this object
        """
        raise NotImplementedError()


class Instance(DICOMObject):
    parent: "Series"

    def all_instances(self):
        """A list containing this instance itself. To match other signatures"""
        return [self]

    def all_series(self) -> List["Series"]:
        return [self.parent]

    def children(self):
        return []

    def root(self):
        return self.parent.parent

    def reference(self) -> InstanceReference:
        """Return a Reference to this object using uids"""
        return InstanceReference(
            study_uid=self.parent.parent.uid,
            series_uid=self.parent.uid,
            instance_uid=self.uid,
        )


class Series(DICOMObject):
    instances: Sequence[Instance]
    parent: "Study"

    def __getitem__(self, instance_uid):
        return self.get(instance_uid)

    def all_instances(self):
        """Each instance contained in this series"""
        return self.instances

    def all_series(self) -> List["Series"]:
        return [self]

    def children(self):
        return self.instances

    def root(self):
        return self.parent

    def get(self, instance_uid: str) -> Instance:
        """Get instance with this uid"""
        try:
            return next(x for x in self.instances if x.uid == instance_uid)
        except StopIteration as e:
            raise KeyError(
                f'instance with uid "{instance_uid}" not found in series'
            ) from e

    def reference(self) -> SeriesReference:
        """Return a Reference to this object using uids"""
        return SeriesReference(study_uid=self.parent.uid, series_uid=self.uid)


class Study(DICOMObject):
    series: Sequence[Series]

    def __getitem__(self, series_uid) -> Series:
        return self.get(series_uid)

    def all_instances(self):
        """Return each instance contained in this study"""
        return list(chain(*(x.instances for x in self.series)))

    def all_series(self) -> List["Series"]:
        return list(self.children())

    def children(self):
        return self.series

    def root(self):
        return self

    def get(self, series_uid: str) -> Series:
        """Get series with this uid"""
        try:
            return next(x for x in self.series if x.uid == series_uid)
        except StopIteration as e:
            raise KeyError(
                f'series with uid "{series_uid}" not found in study'
            ) from e

    def reference(self) -> StudyReference:
        """Return a Reference to this object using uids"""
        return StudyReference(study_uid=self.uid)


class NonInstanceParameterError(DICOMTrolleyError):
    """A DICOMDownloadable could not be converted into its constituent instances.

    Trolley allows the downloading of higher-level DICOM objects like
    download(StudyInstanceUID='1234'). Some Downloader implementations can handle
    this out of the box. Others, like WADO-URI, cannot, and instead require the
    download target to be split into separate instances. This splitting requires
    additional queries, which can only be performed outside the Downloader instance
    itself.
    By raising this error a Downloader method can signal that it cannot process
    the non-instance components of an input parameter. The caller then has an
    opportunity to split the input parameters into instances and try again.

    """


class NonSeriesParameterError(DICOMTrolleyError):
    """A DICOMDownloadable could not be converted into Series or Instance reference.
    Weaker version of NonInstanceParameterError
    """


def to_instance_refs(
    objects: Sequence[DICOMDownloadable],
) -> List[InstanceReference]:
    """Convert all input to instance references. Raise informative errors.

    This is a common pre-processing step used by Downloader classes. Many cannot
    download higher-level objects like 'study 123' directly, but instead require
    you to query all instances contained in 'study 123' first and pass those to
    the downloader.

    Parameters
    ----------
    objects:
        Convert these objects

    Returns
    -------
    List[InstanceReference]
        References to all instances contained in objects

    Raises
    ------
    NonInstanceParameterError
        If any input object could not be converted into instances. This is the
        case for higher-level references like StudyReference, or for DICOMObjects
        that do not contain any deeper-level information, such as a Study that
        contains no Series (which is the correct result of Study-level queries).

    """
    output: List[InstanceReference] = []
    for obj in objects:
        if isinstance(obj, InstanceReference):
            output.append(obj)  # Already an instance reference. Just add
        elif isinstance(obj, DICOMObjectReference):
            raise NonInstanceParameterError(
                f"Cannot extract instances from '{obj}' "
            )
        elif isinstance(obj, DICOMObject):
            instances = obj.all_instances()  # Extract instances
            if instances:
                output = output + [x.reference() for x in instances]
            else:
                raise NonInstanceParameterError(
                    f"{obj} contains no instances. "
                    f"Was this information queried for?"
                )

    return output


def to_series_level_refs(
    objects: Sequence[DICOMDownloadable],
) -> List[Union[SeriesReference, InstanceReference]]:
    """Make sure all input objects are Series or Instance references.
    Convert if possible, raise exceptions if not.

    Weaker version of to_instance_refs(). Weaker because series references without
    instance info are also allowed here. See that function for more info

    Parameters
    ----------
    objects:
        Convert these objects

    Returns
    -------
    List[Union[SeriesReference, InstanceReference]]
        References to series and instances contained in input objects

    Raises
    ------
    NonSeriesParameterError
        If any input object could not be converted into instances. This is the
        case for higher-level references like StudyReference, or for DICOMObjects
        that do not contain any deeper-level information, such as a Study that
        contains no Series (which is the correct result of Study-level queries).

    """
    output: List[Union[InstanceReference, SeriesReference]] = []
    for obj in objects:
        if isinstance(
            obj, (InstanceReference, SeriesReference, Instance, Series)
        ):
            # already OK, just add
            output.append(obj.reference())
        elif isinstance(obj, StudyReference):
            # no series in here, I don't have enough info
            raise NonSeriesParameterError(
                f"Cannot extract series from '{obj}' "
            )
        elif isinstance(obj, Study):
            if obj.series:
                for series in obj.series:
                    output.append(series.reference())
            else:
                raise NonSeriesParameterError(
                    f"Cannot extract series from '{obj}'"
                    f". It contains no series "
                )
    return output


class Downloader:
    """Something that can download DICOM images. Base class"""

    def get_dataset(self, instance: InstanceReference):
        """Get DICOM dataset for the given instance (slice)

        Raises
        ------
        DICOMTrolleyError
            If getting does not work for some reason

        Returns
        -------
        Dataset
            A pydicom dataset
        """
        raise NotImplementedError

    def datasets(self, objects: Sequence[DICOMDownloadable]):
        """Retrieve each instance

        Returns
        -------
        Iterator[Dataset, None, None]

        Raises
        ------
        NonInstanceParameterError
            If objects contain non-instance targets like a StudyInstanceUID and
            download can only process Instance targets. See Exception docstring
            for rationale
        """
        raise NotImplementedError

    def datasets_async(
        self, instances: Sequence[InstanceReference], max_workers=None
    ):
        """Retrieve each instance in multiple threads

        Parameters
        ----------
        instances: Sequence[InstanceReference]
            Retrieve dataset for each of these instances
        max_workers: int, optional
            Use this number of workers in ThreadPoolExecutor. Defaults to
            default for ThreadPoolExecutor

        Raises
        ------
        DICOMTrolleyError
            When a server response cannot be parsed as DICOM

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        raise NotImplementedError


class QueryLevels(str, Enum):
    """Used in dicom queries to indicate how rich the search should be"""

    STUDY = "STUDY"
    SERIES = "SERIES"
    INSTANCE = "INSTANCE"


# Used in Query.init_from_query() type annotation
QuerySubType = TypeVar("QuerySubType", bound="Query")


class Query(BaseModel):
    """A simple DICOM query that is acceptable to all Searcher classes

    Limited parameters but can be used in all Searcher backends

    Notes
    -----
    *  Different backends use slightly different query parameters. This class
       implements all common parameters.

    *  DICOM queries consist of a number of DICOM tags followed by a search
       criterion. The naming format for these parameters follows DICOM conventions
       (CamelCase). In addition, a query has non-DICOM meta-parameters. Here regular
       python (lower_case_underscore) naming is used

    General notes on subclassing Query

    A Searcher class often has its own associated Query subclass. For instance,
    the Mint searcher function Mint.find_study(query) can take a MintQuery instance
    which uniquely allows the setting of the MINT-specific parameter `limit`.
    Any Searcher can however also take a generic Query instance as input. This
    makes it possible to use a simple Query in your code and be independent
    of searcher backend.
    An issue with the scheme is that it is in principle allowed to use the Query
    baseclass as input, it should also be allowed to use any Query subclass in
    any searcher. This, for example, would be legal:

    ```python
    query = QidoRelationalQuery(qido_specific_setting=1)
    studies = Mint().find_study(query)
    ```
    This is weird but not unthinkable. The problem is that the Mint searcher
    cannot do anything with the `qido_specific_setting` set for the query. If
    Mint just uses the basic Query parameters for its query, information would
    be ignored or lost. On the other hand, if we enforce that Mint().find_study()
    only takes MintQuery objects, we lose the ability to use a Query as a
    backend-agnostic general query.

    The tradeoff in structure is between the convenience of having a
    backend-agnostic query on the one hand, and having tight type definitions
    for input arguments on the other.

    In the end, decided that the backend-agnostic query is too useful to lose.
    However, there should be no silent ignoring of parameters. So we will have
    Searcher methods raise UnSupportedParameterError if using an input query would
    result in information being lost.
    """

    # Dicom parameters
    StudyInstanceUID: str = ""
    AccessionNumber: str = ""
    PatientName: str = ""
    PatientID: str = ""
    ModalitiesInStudy: str = ""
    SeriesInstanceUID: str = ""

    # non-DICOM parameters. Translated into derived parameters when querying
    query_level: QueryLevels = (
        QueryLevels.STUDY
    )  # to which depth to return results
    max_study_date: Optional[datetime]
    min_study_date: Optional[datetime]
    include_fields: List[str] = Field([])  #

    class Config:
        extra = "forbid"  # raise ValueError when passing an unknown keyword to init

    @classmethod
    def init_from_query(
        cls: Type[QuerySubType], query: "Query"
    ) -> QuerySubType:
        """Create an instance of this class based on any Query instance

        This function enables Query to be used in all backends. See Query
        notes for rationale

        Parameters
        ----------
        query: Query
            Use all non-falsy parameters from this to create new instance

        Returns
        -------
        An instance of this class based on input query

        Raises
        ------
        UnSupportedParameterError
            If query contains non-default parameters that are not supported in
            this class
        """
        # remove empty, None and 0 values
        params = {key: val for key, val in query.dict().items() if val}
        try:
            return cls(**params)
        except ValidationError as e:
            raise UnSupportedParameterError(
                f"Conversion from {query.to_short_string()} would ignore one more "
                f"parameters. You are probably converting between incompatible "
                f"query subtypes. Use a basic Query instance to avoid this error. "
                f"See ValidationError for details on which parameter is causing "
                f"the problem"
            ) from e

    @staticmethod
    def validate_keyword(keyword):
        if not tag_for_keyword(keyword):
            raise ValueError(f"{keyword} is not a valid DICOM keyword")

    @validator("include_fields")
    def include_fields_check(cls, include_fields, values):  # noqa: B902, N805
        """Include fields should be valid dicom tag names"""
        for field in include_fields:
            if not tag_for_keyword(field):
                raise ValueError(f"{field} is not a valid DICOM keyword")

        return include_fields

    def to_short_string(self):
        """A more information-dense str repr. For human reading"""
        filled_fields = {key: val for key, val in self.dict().items() if val}
        filled_fields["query_level"] = filled_fields["query_level"].value
        return f"{type(self).__name__}: {filled_fields}"


class ExtendedQuery(Query):
    """Query with additional parameters. Base for Mint and DICOM-QR.

    Made this to avoid duplicate extra fields in Mint and DICOM-QR searchers
    """

    InstitutionName: str = ""
    PatientSex: str = ""
    StudyDescription: str = ""
    SeriesDescription: str = ""
    InstitutionalDepartmentName: str = ""
    PatientBirthDate: Optional[date]


class Searcher:
    """Something that can search for DICOM studies. Base class."""

    def find_studies(self, query: Query) -> Sequence[Study]:
        raise NotImplementedError()

    def find_study(self, query: Query) -> Study:
        """Like find_studies, but guarantees exactly one result. Exception if not.

        This method is meant for searches that contain unique identifiers like
        StudyInstanceUID, AccessionNumber, etc.

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        results = self.find_studies(query)
        if len(results) == 0 or len(results) > 1:
            raise DICOMTrolleyError(
                f"Expected exactly one study for query '{query.to_short_string()}',"
                f" but found {len(results)}"
            )
        return results[0]

    def find_full_study_by_id(self, study_uid: str) -> Study:
        """Find a single study at image level

        Useful for automatically finding all instances for a study. Meant to be
        implemented in child classes

        Parameters
        ----------
        study_uid: str
            Study to search for

        Returns
        -------
        Study

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        raise NotImplementedError()

    def find_study_at_instance_level(self, study_uid: str) -> Study:
        """Find a single study at image level

        Useful for automatically finding all instances for a study.

        Parameters
        ----------
        study_uid: str
            Study to search for

        Returns
        -------
        Study
            Containing full DICOM object information, series and instances

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        return self.find_study(
            Query(StudyInstanceUID=study_uid, query_level=QueryLevels.INSTANCE)
        )

    def find_study_at_series_level(self, study_uid: str) -> Study:
        """Find a single study at series level

        Meant to be
        implemented in child classes

        Parameters
        ----------
        study_uid: str
            Study to search for

        Returns
        -------
        Study
            Containing series, but not instances

        Raises
        ------
        DICOMTrolleyError
            If no results or more than one result is returned by query
        """
        return self.find_study(
            Query(StudyInstanceUID=study_uid, query_level=QueryLevels.SERIES)
        )


Instance.update_forward_refs()  # enables pydantic validation
Series.update_forward_refs()



"""Models WADO-RS: Web Access to dicom Objects by Restful Services

[Wado RS description]
(https://www.dicomstandard.org/using/dicomweb/retrieve-wado-rs-and-wado-uri/)

See Also
--------
[DICOM part18 section 10.4]
(https://dicom.nema.org/medical/dicom/current/output/chtml/part18/sect_10.4.html)

Notes
-----
Models only the parts of WADO-RS directly related to downloading DICOM image data.
WADO-RS also supports downloading metadata and rendered images, but these are
outside the scope of the dicomtrolley project.

Specifically, from DICOM PS3.18 section 10.4

Download supported by dicomtrolley:

* Instance resources (download all instances)

Download Not Supported by dicomtrolley:

* Metadata resources

* Rendered resources

* Thumbnail resources

* Bulkdata resources

* Pixel Data resources

"""
from itertools import chain
from typing import Iterator, Sequence

from pydicom import Dataset, dcmread
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomBytesIO

from dicomtrolley.core import (
    DICOMDownloadable,
    DICOMObjectReference,
    Downloader,
    InstanceReference,
    SeriesReference,
    StudyReference,
    to_series_level_refs,
)
from dicomtrolley.exceptions import DICOMTrolleyError
from dicomtrolley.http import HTTPMultiPartStream
from dicomtrolley.logs import get_module_logger

logger = get_module_logger("wado_rs")


class WadoRS(Downloader):
    """A connection to a WADO-RS server"""

    def __init__(
        self, session, url, http_chunk_size=5242880, request_per_series=True
    ):
        """
        Parameters
        ----------
        session: requests.session
            A logged-in session over which WADO calls can be made
        url: str
            WADO-RS endpoint, including protocol and port. Like
            https://server:8080/wado
        http_chunk_size: int, optional
            Number of bytes to read each time when streaming chunked responses.
            Defaults to 5MB (5242880 bytes)
        request_per_series: bool, optional
            If true, split requests per series when downloading. If false,
            request all instances at once. Splitting reduces load on server.
            defaults to True.
        """

        self.session = session
        self.url = url
        self.http_chunk_size = http_chunk_size
        self.request_per_series = request_per_series

    def datasets(self, objects: Sequence[DICOMDownloadable]):
        """Retrieve each instance

        Returns
        -------
        Iterator[Dataset, None, None]

        Raises
        ------
        NonSeriesParameterError
            If request_per_series is True and objects contains study references
            without series information.
        """
        logger.debug("Getting datasets")
        if isinstance(objects, DICOMDownloadable):
            objects = [objects]  # handle passing single object instead of list

        if self.request_per_series:
            references: Sequence[DICOMDownloadable] = to_series_level_refs(
                objects
            )
            logger.debug(
                f"Splitting {len(objects)} objects into series. After split,"
                f" getting {len(references)} downloadables"
            )
        else:
            references = objects

        return chain.from_iterable(
            self.download_iterator(obj) for obj in references
        )

    def download_iterator(self, downloadable: DICOMDownloadable):
        """Perform a wado RS request and iterate over the returned datasets

        Returns
        -------
        Iterator[Dataset, None, None]
            All datasets included in the response
        """
        uri = self.wado_rs_instance_uri(downloadable.reference())
        logger.debug(f"Calling {uri}")
        response = self.session.get(
            url=uri,
            stream=True,
        )

        return self.parse(response)

    def parse(self, response) -> Iterator[Dataset]:
        """Extract datasets out of http response from a rad69 server

        Parameters
        ----------
        response:
            A requests response objects, requests with stream=True

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected or if parsing fails

        Returns
        -------
        Iterator[Dataset, None, None]
            All datasets included in this response
        """

        logger.debug("Parsing WADO-RS response")

        self.check_for_response_errors(response)

        part_stream = HTTPMultiPartStream(
            response, stream_chunk_size=self.http_chunk_size
        )
        for part in part_stream:
            raw = DicomBytesIO(part.content)
            try:
                yield dcmread(raw)
            except InvalidDicomError as e:
                raise DICOMTrolleyError(
                    f"Error parsing response as dicom: {e}."
                    f" Response content (first 300 elements) was"
                    f" {str(response.content[:300])}"
                ) from e

    @staticmethod
    def check_for_response_errors(response):
        """Raise exceptions if this response is not a valid WADO-RS response.

        Parameters
        ----------
        response: response
            response as returned from a wado-rs call

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected
        """
        if response.status_code != 200:
            raise DICOMTrolleyError(
                f"Calling {response.url} failed ({response.status_code} - "
                f"{response.reason})\n"
                f"response content was {str(response.content[:300])}"
            )

        # check multipart
        if "Content-type" not in response.headers:
            raise DICOMTrolleyError(
                f"Expected multipart response, but got no content type for this"
                f" response. Start of response: {str(response.content[:300])}"
            )

    def wado_rs_instance_uri(self, reference: DICOMObjectReference):
        """WADO-RS URI to request all instances contained in referenced object"""
        uri = self.url.rstrip(
            "/"
        )  # self.url might or might not have trailing /
        if isinstance(reference, StudyReference):
            return f"{uri}/studies/{reference.study_uid}"
        elif isinstance(reference, SeriesReference):
            return (
                f"{uri}/studies/{reference.study_uid}/series"
                f"/{reference.series_uid}"
            )
        elif isinstance(reference, InstanceReference):
            return (
                f"{uri}/studies/{reference.study_uid}/series"
                f"/{reference.series_uid}/instances/{reference.instance_uid}"
            )



"""Combines Searcher and Downloader to make getting DICOM studies easy.

Notes
-----
Design choices:

Searcher and Downloader classes should be stand-alone. They are not allowed to
communicate directly. Trolley has knowledge of both and is in control.
"""
import itertools
import tempfile
from typing import List, Optional, Sequence, Union

from dicomtrolley.core import (
    DICOMDownloadable,
    DICOMObject,
    DICOMObjectReference,
    Downloader,
    Instance,
    InstanceReference,
    NonInstanceParameterError,
    NonSeriesParameterError,
    Searcher,
    Series,
    SeriesReference,
    Study,
)
from dicomtrolley.exceptions import DICOMTrolleyError
from dicomtrolley.logs import get_module_logger
from dicomtrolley.parsing import DICOMObjectNotFound, DICOMObjectTree
from dicomtrolley.storage import DICOMDiskStorage, StorageDir


logger = get_module_logger("trolley")


class Trolley:
    """Combines a search and download method to get DICOM studies easily

    Features:

    * Searching for DICOM using a Query instance is backend-agnostic.

    * If a download method requires additional information such as all instance UIDs,
      trolley can query for these in the background.

    * Saves to disk in reasonable (uid based) folder structure.
    """

    def __init__(
        self,
        downloader: Downloader,
        searcher: Searcher,
        query_missing=True,
        storage: Optional[DICOMDiskStorage] = None,
    ):
        """

        Parameters
        ----------
        downloader: Downloader
            The module to use for downloads
        searcher: Searcher
            The module to use for queries
        query_missing: bool, optional
            if True, Trolley.download() will query for missing DICOM instances. For
            example when passing a Study obtained from a study-level query, which does
            not contain any information on instances
            if False, missing instances will raise MissingObjectInformationError
        storage: DICOMDiskStorage instance, optional
            All downloads are saved to disk by calling this objects' save() method.
            Defaults to basic StorageDir (saves as /studyid/seriesid/instanceid)
        """
        self.downloader = downloader
        self.searcher = searcher
        self._searcher_cache = CachedSearcher(
            searcher, query_missing=query_missing
        )
        if storage:
            self.storage = storage
        else:
            self.storage = StorageDir(tempfile.gettempdir())

    def find_studies(self, query) -> List[Study]:
        """Find study information

        Parameters
        ----------
        query:
            Search for these criteria

        Returns
        -------
        List[Study]
        """
        return list(self.searcher.find_studies(query))

    def find_study(self, query) -> Study:
        """Like find study, but returns exactly one result or raises exception.

        For queries using study identifiers such as StudyInstanceUID,AccessionNumber

        Parameters
        ----------
        query
            Search for these criteria. Query documentation for more info.

        Returns
        -------
        Study
        """
        return self.searcher.find_study(query)

    def download(
        self,
        objects: Union[DICOMDownloadable, Sequence[DICOMDownloadable]],
        output_dir,
        use_async=False,
        max_workers=None,
    ):
        """Download the given objects to output dir."""
        if not isinstance(objects, Sequence):
            objects = [objects]  # if just a single item to download is passed
        logger.info(f"Downloading {len(objects)} object(s) to '{output_dir}'")
        if use_async:
            datasets = self.fetch_all_datasets_async(
                objects=objects, max_workers=max_workers
            )
        else:
            datasets = self.fetch_all_datasets(objects=objects)

        for dataset in datasets:
            self.storage.save(dataset=dataset, path=output_dir)

    def fetch_all_datasets(self, objects: Sequence[DICOMDownloadable]):
        """Get full DICOM dataset for all instances contained in objects.

        Returns
        -------
        Iterator[Dataset]
            The downloaded dataset and the context that was used to download it
        """
        try:
            yield from self.downloader.datasets(objects)
        except NonSeriesParameterError:
            # downloader wants at least series level information. Do extra work.
            series_lvl_objects = self.ensure_to_series_level(objects)
            yield from self.downloader.datasets(series_lvl_objects)
        except NonInstanceParameterError:
            # downloader wants only instance input. Do extra work.
            instances = self.convert_to_instances(objects)
            yield from self.downloader.datasets(instances)

    def convert_to_instances(
        self, objects_in: Sequence[DICOMDownloadable]
    ) -> List[InstanceReference]:
        """Find all instances contained in objects, running additional image-level
        queries with searcher if needed.

        Holds DICOMObjects in an internal cache, avoiding queries as much as
        possible.

        Note
        ----
        This method can fire additional search queries and might take a long time to
        return depending on the number of objects and missing instances
        """
        return list(
            itertools.chain.from_iterable(
                self._searcher_cache.retrieve_instance_references(x)
                for x in objects_in
            )
        )

    def ensure_to_series_level(
        self, objects_in: Sequence[DICOMDownloadable]
    ) -> Sequence[Union[InstanceReference, SeriesReference]]:
        """Make sure all input is converted to instance or series

        ----
        This method can fire additional search queries and might take a long time to
        return depending on the number of objects and missing instances or series
        """
        return list(
            itertools.chain.from_iterable(
                self._searcher_cache.ensure_series_level_references(x)
                for x in objects_in
            )
        )

    def fetch_all_datasets_async(self, objects, max_workers=None):
        """Get DICOM dataset for each instance given objects using multiple threads.

        Parameters
        ----------
        objects: Sequence[DICOMDownloadable]
            get dataset for each instance contained in these objects
        max_workers: int, optional
            Max number of ThreadPoolExecutor workers to use. Defaults to
            ThreadPoolExecutor default

        Raises
        ------
        DICOMTrolleyError
            If getting or parsing of any instance fails

        Returns
        -------
        Iterator[Dataset, None, None]
            The downloaded dataset and the context that was used to download it
        """
        try:
            yield from self.downloader.datasets_async(
                instances=objects,
                max_workers=max_workers,
            )
        except NonInstanceParameterError:
            yield from self.downloader.datasets_async(
                instances=self.convert_to_instances(objects),
                max_workers=max_workers,
            )


class CachedSearcher:
    def __init__(
        self,
        searcher: Searcher,
        cache: Optional[DICOMObjectTree] = None,
        query_missing: Optional[bool] = True,
    ):
        """A DICOMObject tree (study/series/instances) that will launch
        queries to expand itself if needed. Tries to query as little as
        possible.

        Created this to efficiently get all instances for download based on a
        variable collection of DICOMDownloadable objects.

        Parameters
        ----------
        searcher: Searcher
            Use this searcher to search for missing elements in cache
        cache: DICOMObjectTree, Optional
            Use this tree as cache. Defaults to an empty tree
        query_missing: Bool, optional
            Launch queries to find missing information. If False, will raise
            MissingObjectInformationError. Defaults to True
        """
        self.searcher = searcher
        if not cache:
            cache = DICOMObjectTree([])
        self.cache = cache
        self.query_missing = query_missing

    def retrieve_instance_references(
        self, downloadable: DICOMDownloadable
    ) -> List[InstanceReference]:
        """Get references for all instances contained in downloadable, performing
        additional queries if needed

        Raises
        ------
        MissingObjectInformationError
            If Trolley.query_missing is False and queries would be required to
            retrieve all instances.

        """
        # DICOMObject might already contain all required info
        if isinstance(downloadable, DICOMObject):
            instances = downloadable.all_instances()
            if instances:
                return [x.reference() for x in instances]

        # No instances in downloadable. Do we have instances in cache for this?
        reference = downloadable.reference()
        try:
            return [
                x.reference() for x in self.get_instances_from_cache(reference)
            ]
        except MissingObjectInformationError as e:  # not found, we'll have to query
            if not self.query_missing:
                raise MissingObjectInformationError(
                    f"No instances cached for {reference} "
                    f"and query_missing was False"
                ) from e

            logger.debug(f"No instances cached for {reference}. Querying")
            self.query_study_instances(reference)
            return [
                x.reference() for x in self.get_instances_from_cache(reference)
            ]

    def ensure_series_level_references(
        self, downloadable: DICOMDownloadable
    ) -> List[Union[InstanceReference, SeriesReference]]:
        """Extract references of at least series level. Query if not found

        Raises
        ------
        MissingObjectInformationError
            If Trolley.query_missing is False and queries would be required to
            retrieve Series.


        TODO: Strong whiffs of code smell here. This method should not be this
        long. Some pointers for refactoring: Get rid of explicit type checking.
        possibly move logic to DICOMDownloadable classes. Why does this method
        take a DICOMDownloadable and not just a reference? It seems conversion logic
        is shoehorned into CachedSearcher and Trolley classes. CachedSearcher should
        govern a cache of DICOM information, Trolley should handle high level
        requests. Neither are quite right for conversion. Possibly create separate
        class for this.
        """
        # Instance and Series should just be left as is
        if isinstance(
            downloadable,
            (Instance, Series, InstanceReference, SeriesReference),
        ):
            return [downloadable.reference()]
        # If its a study, it might need work

        elif isinstance(downloadable, Study):
            series: Sequence[
                Union[Instance, Series]
            ] = downloadable.all_series()
            if series:  # there were series inside Study already
                return [x.reference() for x in series]

        # no studies. Maybe they are in cache?
        reference = downloadable.reference()
        try:
            series = self.get_series_level_from_cache(reference)
            if series:
                return [x.reference() for x in series]
            else:
                raise MissingObjectInformationError()  # not expected. Being careful.

        except MissingObjectInformationError as e:  # not found, we'll have to query
            if not self.query_missing:
                raise MissingObjectInformationError(
                    f"No series cached for {reference} "
                    f"and query_missing was False"
                ) from e

            logger.debug(f"No series cached for {reference}. Querying")
            self.query_study_series(reference)
            return [
                x.reference()
                for x in self.get_series_level_from_cache(reference)
            ]

    def get_instances_from_cache(
        self, reference: DICOMObjectReference
    ) -> List[Instance]:
        """Retrieve all instances for this reference. Raise exception if not found

        Raises
        ------
        MissingObjectInformationError
            If no instances are found in cache
        """
        try:
            instances = self.cache.retrieve(reference).all_instances()
            if instances:
                return instances
            else:
                # reference was in cache, but there are no instances
                raise MissingObjectInformationError(
                    f"No instances found for {reference}"
                )
        except DICOMObjectNotFound as e:
            # reference was not in cache
            raise MissingObjectInformationError(
                f"No instances found for {reference}"
            ) from e

    def get_series_level_from_cache(
        self, reference: DICOMObjectReference
    ) -> Sequence[Union[Instance, Series]]:
        """Retrieve series this reference, or instance if required.
        Raise exception if not found

        Raises
        ------
        MissingObjectInformationError
            If no series or instance can be found for this reference
        """
        try:
            series: Sequence[Union[Instance, Series]] = self.cache.retrieve(
                reference
            ).all_series()
            if series:
                return series
            else:
                # reference was in cache, but there are no instances
                raise MissingObjectInformationError(
                    f"No series found for {reference}"
                )
        except DICOMObjectNotFound as e:
            # reference was not in cache
            raise MissingObjectInformationError(
                f"No series found for {reference}"
            ) from e

    def query_study_instances(self, reference: DICOMObjectReference):
        study = self.searcher.find_study_at_instance_level(reference.study_uid)
        self.cache.add_study(study)

    def query_study_series(self, reference: DICOMObjectReference):
        study = self.searcher.find_study_at_series_level(reference.study_uid)
        self.cache.add_study(study)


class MissingObjectInformationError(DICOMTrolleyError):
    """An operation on a DICOM object cannot complete because the required information
    is not available. For example, trying to extract all instances from a Study that
    was retrieved without instance information
    """

    pass



"""Models QIDO-RS: Query based on ID for dicom Objects by Restful Services.


See: [QIDO-RS reference](https://www.dicomstandard.org/using/dicomweb/query-qido-rs/)
and [QIDO-RS at NEMA]
(https://dicom.nema.org/medical/dicom/current/output/chtml/part18/
sect_10.6.html#sect_10.6.1.2)

"""
import json
from datetime import datetime
from typing import Dict, List, Optional, Sequence, Union

from pydantic import root_validator
from pydicom import Dataset
from requests import Response

from dicomtrolley.core import (
    Query,
    QueryLevels,
    Searcher,
    Study,
)
from dicomtrolley.exceptions import (
    DICOMTrolleyError,
    UnSupportedParameterError,
)
from dicomtrolley.logs import get_module_logger
from dicomtrolley.parsing import DICOMParseTree

logger = get_module_logger("qido_rs")


class QidoRSQueryBase(Query):
    """Base query class as defined in DICOM PS3.18 2023b section 8.3.4
    table 8.3.4-1
    """

    limit: int = 0  # How many results to return. 0 = all
    offset: int = 0  # Number of skipped results

    @root_validator()  # type: ignore
    def min_max_study_date_xor(cls, values):  # noqa: B902, N805
        """Min and max should both be given or both be empty"""
        min_date = values.get("min_study_date")
        max_date = values.get("max_study_date")
        if min_date and not max_date:
            raise ValueError(
                f"min_study_date parameter was passed"
                f"({min_date}), "
                f"but max_study_date was not. Both need to be given"
            )
        elif max_date and not min_date:
            raise ValueError(
                f"max_study_date parameter was passed ({max_date}), "
                f"but min_study_date was not. Both need to be given"
            )
        return values

    @staticmethod
    def date_to_str(date_in: Optional[datetime]) -> str:
        """Date to WADO-RS URI format"""
        if not date_in:
            return ""
        return date_in.strftime("%Y%m%d")

    @staticmethod
    def date_range_to_str(
        min_date: Optional[datetime], max_date: Optional[datetime]
    ) -> str:
        """String indicating a date or time range following
        https://dicom.nema.org/medical/dicom/current/output/
        chtml/part04/sect_C.2.2.2.5.html

        Notes
        -----
        For wado-rs, both StudyDate=20001012 and StudyDate=20001012-20001012
        denote the same single day. To keep thing simple this function does not
        collapse 20001012-20001012, just leaves it as is
        """
        if not min_date and not max_date:
            raise ValueError("Cannot create a date range without any dates")
        to_str = QidoRSQueryBase.date_to_str

        return f"{to_str(min_date)}-{to_str(max_date)}"

    def uri_base(self) -> str:
        """WADO-RS url to call when performing this query. Full URI also needs
        uri_search_params()

        The non-query part of the URI as defined in
        DICOM PS3.18 section 10.6 table 10.6.1-2
        """
        raise NotImplementedError

    def uri_search_params(self) -> Dict[str, Union[str, List[str]]]:
        """The search parameter part of the URI as defined in
        DICOM PS3.18 section 10.6 table 10.6.1-2

        Returns
        -------
        Dict[str, Union[str, List[str]]]
            Output that can be fed directly into a requests post request.
            format is param_name:param_value. If param_value is a list, the param
            will be included multiple times.
            See https://docs.python-requests.org/en/latest/user/quickstart/
            #passing-parameters-in-urls

        Notes
        -----
        Will not output any parameters with Null or empty value (bool(value)==False).
        This does not affect query functionality but makes output cleaner in strings
        """
        search_params: Dict[str, Union[str, List[str]]] = {}

        # parse dates
        if self.min_study_date or self.max_study_date:
            search_params["StudyDate"] = self.date_range_to_str(
                self.min_study_date, self.max_study_date
            )
        # parse include fields
        if self.include_fields:
            search_params["includefield"] = self.include_fields

        # now collect all other Query() fields that can be search params
        exclude_fields = {
            "min_study_date",  # addressed above
            "max_study_date",
            "include_fields",
            "StudyInstanceUID",  # potential part of url, not search params
            "SeriesInstanceUID",
            "SOPClassInstanceUID",
            "query_level",
        }

        other_search_params = {
            key: val
            for key, val in self.dict().items()
            if key not in exclude_fields
        }

        search_params.update(other_search_params)
        return {
            key: val for key, val in search_params.items() if val
        }  # remove empty


class HierarchicalQuery(QidoRSQueryBase):
    """QIDO-RS Query that uses that traditional study->series->instance structure

    Allows the following queries:

    * All Studies

    * Study's Series

    * Study's Series' Instances

    See https://dicom.nema.org/medical/dicom/current/output/html/part18.html#sect_10.6

    Faster than relationalQuery, but requires more information
    """

    @root_validator()  # type: ignore
    def uids_should_be_hierarchical(cls, values):  # noqa: B902, N805
        """Any object uids passed should conform to study->series->instance"""
        order = ["StudyInstanceUID", "SeriesInstanceUID", "SOPInstanceUID"]

        def assert_parents_filled(a_hierarchy, value_dict):
            """Assert that if a value in hierarchy is filled, its parent is
            filled too
            """
            if len(a_hierarchy) <= 1:  # if there is only one item, this
                return True  # is either fine, or checked in last iteration
            current = a_hierarchy.pop()
            value = value_dict.get(current)
            parent = value_dict.get(a_hierarchy[-1])
            if value and not parent:  # all parents should be filled then
                raise ValueError(
                    f"This query is not hierarchical. {current} "
                    f"(value:{value})is given , but parent, "
                    f"{a_hierarchy[-1]}, is not. Add parent IDs or "
                    f"use a relational Query instead"
                )
            else:
                return assert_parents_filled(a_hierarchy, value_dict)

        assert_parents_filled(order, values)

        return values

    @root_validator()  # type: ignore
    def uids_should_match_query_level(cls, values):  # noqa: B902, N805
        """If a query is for instance level, there should be study and series UIDs"""
        query_level = values["query_level"]

        def assert_key_exists(values_in, query_level_in, missing_key_in):
            if not values_in.get(missing_key_in):
                raise ValueError(
                    f'To search at query level "{query_level_in}" '
                    f"you need to supply a {missing_key_in}. Or use "
                    f"a QIDO-RS relational query"
                )

        if query_level == QueryLevels.STUDY:
            pass  # Fine. you can always look for some studies
        elif query_level == QueryLevels.SERIES:
            assert_key_exists(values, query_level, "StudyInstanceUID")
        elif query_level == QueryLevels.INSTANCE:
            assert_key_exists(values, query_level, "SeriesInstanceUID")
            assert_key_exists(values, query_level, "StudyInstanceUID")

        return values

    def uri_base(self) -> str:
        """WADO-RS url to call when performing this query. Full URI also needs
        uri_search_params()

        The non-query part of the URI as defined in
        DICOM PS3.18 section 10.6 table 10.6.1-2
        """

        if self.query_level == QueryLevels.STUDY:
            return "/studies"
        elif self.query_level == QueryLevels.SERIES:
            return f"/studies/{self.StudyInstanceUID}/series"
        elif self.query_level == QueryLevels.INSTANCE:
            return (
                f"/studies/{self.StudyInstanceUID}/series/"
                f"{self.SeriesInstanceUID}/instances"
            )
        else:
            raise ValueError(
                f'Unknown querylevel "{self.query_level}". '
                f'Should be one of "{QueryLevels}"'
            )


# used in RelationalQuery
STUDY_VALUE_ERROR_TEXT: str = (
    "Query level STUDY makes no sense for a QIDO-RS "
    "relational query. Perhaps use a hierarchical query?"
)


class RelationalQuery(QidoRSQueryBase):
    """QIDO-RS query that allows querying for series and instances directly

    Allows the following queries

    * Study's Instances

    * All Series

    * All Instances

    See https://dicom.nema.org/medical/dicom/current/output/html/part18.html#sect_10.6

    Allows broader searches than HierarchicalQuery, but can be slower
    """

    @root_validator()  # type: ignore
    def query_level_should_be_series_or_instance(
        cls, values  # noqa: B902, N805
    ):
        """A relational query only makes sense for the instance and series levels.
        If you want to look for studies, us a hierarchical query
        """
        if values["query_level"] == QueryLevels.STUDY:
            raise ValueError(STUDY_VALUE_ERROR_TEXT)

        return values

    def uri_base(self) -> str:
        """WADO-RS url to call when performing this query. Full URI also needs
        uri_search_params()

        The non-query part of the URI as defined in
        DICOM PS3.18 section 10.6 table 10.6.1-2

        Raises
        ------
        ValueError
            if query_level is STUDY. This makes no sense for a relational query
        """

        if self.query_level == QueryLevels.STUDY:
            raise ValueError(STUDY_VALUE_ERROR_TEXT)
        elif self.query_level == QueryLevels.SERIES:
            return "/studies"
        elif self.query_level == QueryLevels.INSTANCE:
            if self.StudyInstanceUID:
                # all instances for this study
                return f"/studies/{self.StudyInstanceUID}/instances"
            else:
                # all instances on the intire server (might be slow)
                return "/instances"

        else:
            raise ValueError(
                f'Unknown querylevel "{self.query_level}". '
                f'Should be one of "{QueryLevels}"'
            )


class QidoRS(Searcher):
    """A connection to a QIDO-RS server"""

    def __init__(self, session, url):
        """
        Parameters
        ----------
        session: requests.session
            A logged-in session over which WADO calls can be made
        url: str
            QIDO-RS endpoint, including protocol and port. Like
            https://server:8080/qido
        """

        self.session = session
        self.url = url

    @staticmethod
    def check_for_response_errors(response):
        """Raise exceptions if this response is not a valid WADO-RS response.

        Parameters
        ----------
        response: Response
            requests.Response as returned from a wado-rs call

        Raises
        ------
        NoQueryResults
            If http 204 (No Content) is returned by server
            see https://dicom.nema.org/medical/dicom/current/output/chtml/
            part18/sect_8.3.4.4.html
        DICOMTrolleyError
            If response is otherwise not as expected
        """
        if response.status_code == 204:
            raise NoQueryResults("Server returned http 204 (No Content)")
        elif response.status_code != 200:
            raise DICOMTrolleyError(
                f"Calling {response.url} failed ({response.status_code} - "
                f"{response.reason})\n"
                f"response content was {str(response.content[:300])}"
            )

    @classmethod
    def ensure_query_type(cls, query: Query) -> QidoRSQueryBase:
        """Make sure query is of a type usable in this searcher. Cast if needed

        Separate casting method needed in addition to Query.init_from_query()
        To properly handle the two QIDO-RS query types
        """
        if isinstance(query, QidoRSQueryBase):
            return query  # no conversion, just us whatever it was
        elif isinstance(query, Query):
            # We need to convert. Hierarchical is faster and more straightforward
            try:
                logger.debug(
                    "qido-rs searcher got plain Query object. Converting "
                    "to HierarchicalQuery"
                )
                return HierarchicalQuery.init_from_query(query)
            except UnSupportedParameterError:
                logger.debug(
                    "Converting to HierarchicalQuery did not work. "
                    "Trying slower but less stringent RelationalQuery"
                )
                return RelationalQuery.init_from_query(query)
        else:
            raise ValueError(
                f'Invalid query. Expecting Query, but got "{type(query)}")'
            )

    def find_studies(self, query: Query) -> Sequence[Study]:
        logger.debug(f"Firing query {query.to_short_string()}")

        query = self.ensure_query_type(query)
        url = self.url.rstrip("/") + query.uri_base()
        response = self.session.get(url=url, params=query.uri_search_params())

        try:
            self.check_for_response_errors(response)
        except NoQueryResults:
            return []
        return self.parse_qido_response(json.loads(response.text))

    @staticmethod
    def parse_qido_response(response: Response) -> List[Study]:
        """Assumes response has been json-decoded

        response could contain instances, series or studies
        """
        tree = DICOMParseTree()
        for item in response:
            tree.insert_dataset(Dataset.from_json(item))

        return tree.as_studies()


class NoQueryResults(DICOMTrolleyError):
    """Raised when a query returns 0 results"""



"""Retrieve data from servers using the rad69 protocol

see [rad69 document set]
(https://gazelle.ihe.net/content/rad-69-retrieve-imaging-document-set)
And the corresponding [transaction]
(https://profiles.ihe.net/ITI/TF/Volume2/ITI-43.html)
"""

import math
import uuid
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from itertools import chain
from typing import Dict, List, Sequence
from xml.etree import ElementTree

from jinja2.environment import Template
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomBytesIO
from pydicom.filereader import dcmread
from requests_futures.sessions import FuturesSession

from dicomtrolley.core import (
    DICOMDownloadable,
    Downloader,
    InstanceReference,
    to_instance_refs,
)
from dicomtrolley.exceptions import DICOMTrolleyError
from dicomtrolley.http import HTTPMultiPartStream
from dicomtrolley.logs import get_module_logger
from dicomtrolley.parsing import DICOMParseTree
from dicomtrolley.xml_templates import (
    RAD69_SOAP_REQUEST_TEMPLATE,
    RAD69_SOAP_RESPONSE_ERROR_XPATH,
)

logger = get_module_logger("rad69")


class Rad69(Downloader):
    """A connection to a Rad69 server"""

    def __init__(
        self,
        session,
        url,
        http_chunk_size=5242880,
        request_per_series=True,
        errors_to_ignore=None,
    ):
        """
        Parameters
        ----------
        session: requests.session
            A logged in session over which rad69 calls can be made
        url: str
            rad69 endpoint, including protocol and port. Like https://server:2525/rids
        http_chunk_size: int, optional
            Number of bytes to read each time when streaming chunked rad69 responses.
            Defaults to 5MB (5242880 bytes)
        request_per_series: bool, optional
            If true, split rad69 requests per series when downloading. If false,
            request all instances at once. Splitting reduces load on server.
            defaults to True.
        errors_to_ignore: List[Type], optional
            Errors of this type encountered during download are caught and skipped.
            Defaults to empty list, meaning any error is propagated
        """

        self.session = session
        self.url = url
        self.http_chunk_size = http_chunk_size
        if errors_to_ignore is None:
            errors_to_ignore = []
        self.errors_to_ignore = errors_to_ignore
        self.template = RAD69_SOAP_REQUEST_TEMPLATE
        self.post_headers = {"Content-Type": "application/soap+xml"}
        self.request_per_series = request_per_series

    def datasets(self, objects: Sequence[DICOMDownloadable]):
        """Retrieve all instances via rad69

        A Rad69 request typically contains multiple instances. The data for all
        instances is then streamed back as one multipart http response

        Raises
        ------
        NonInstanceParameterError
            If objects contain non-instance targets like a StudyInstanceUID.
            Rad69 can only download instances

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        instances = to_instance_refs(objects)  # raise exception if needed
        logger.info(f"Downloading {len(instances)} instances")
        if self.request_per_series:
            per_series: Dict[str, List[InstanceReference]] = defaultdict(list)
            for x in instances:
                per_series[x.series_uid].append(x)
            logger.info(
                f"Splitting per series. Found {len(per_series)} series"
            )
            return chain.from_iterable(
                self.series_download_iterator(x, index)
                for index, x in enumerate(per_series.values())
            )

        else:
            return self.download_iterator(instances)

    def series_download_iterator(
        self, instances: Sequence[InstanceReference], index=0
    ):
        """Identical to create_download_iterator, except adds a debug log call"""
        if instances:
            logger.debug(
                f"Downloading series {index}: " f"{instances[0].series_uid}"
            )
        return self.download_iterator(instances)

    def download_iterator(self, instances: Sequence[InstanceReference]):
        """Perform a rad69 request and iterate over the returned datasets

        Returns
        -------
        Iterator[Dataset, None, None]
            All datasets included in the response
        """
        response = self.session.post(
            url=self.url,
            headers=self.post_headers,
            data=self.create_instances_request(instances),
            stream=True,
        )

        return self.parse_rad69_response(response)

    def create_instance_request(self, instance: InstanceReference):
        """Create the SOAP xml structure needed to request an instance from a rad69
        server
        """
        return self.create_instances_request(instances=[instance])

    def create_instances_request(self, instances: Sequence[InstanceReference]):
        """Create the SOAP xml structure to request all given instances from server"""
        # Turn instanceReference list back into study level
        tree = DICOMParseTree()
        for instance in instances:
            tree.insert(
                data=[],
                study_uid=instance.study_uid,
                series_uid=instance.series_uid,
                instance_uid=instance.instance_uid,
            )
        studies = tree.as_studies()

        return Template(self.template).render(
            uuid=str(uuid.uuid4()),
            studies=studies,
            transfer_syntax_list=[
                "1.2.840.10008.1.2.4.70",
                "1.2.840.10008.1.2",
                "1.2.840.10008.1.2.1",
            ],
        )

    def get_dataset(self, instance: InstanceReference):
        """Get DICOM dataset for the given instance (slice)

        Raises
        ------
        DICOMTrolleyError
            If getting does not work for some reason

        Note
        ----
        Getting single datasets is not efficient in rad69. If you want to loop
        this function, use Rad69.datasets() instead

        Returns
        -------
        Dataset
            A pydicom dataset
        """

        response = self.session.post(
            url=self.url,
            headers=self.post_headers,
            data=self.create_instance_request(instance),
        )
        return list(self.parse_rad69_response(response))[0]

    def verify_response(self, response):
        """Check for errors in rad69 response and handle them"""

    @staticmethod
    def check_for_response_errors(response):
        """Raise exceptions if this response is not a multi-part rad69 soap response.

        Parameters
        ----------
        response: response
            response as returned from a rad69 call

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected
        """
        if response.status_code != 200:
            raise DICOMTrolleyError(
                f"Calling {response.url} failed ({response.status_code} - "
                f"{response.reason})\n"
                f"response content was {str(response.content[:300])}"
            )

        # check multipart
        if "Content-type" not in response.headers:
            raise DICOMTrolleyError(
                f"Expected multipart response, but got no content type for this"
                f" response. Start of response: {str(response.content[:300])}"
            )

        if "multipart" not in response.headers["Content-Type"].lower():
            # Maybe server is sending back a valid soap error document.
            Rad69.parse_rad69_soap_error_response(response)

    @staticmethod
    def parse_rad69_soap_error_response(response):
        """Interpret response as rad69 error, raise more helpful exceptions

        Raises
        ------
        XDSMissingDocumentError:
            If data for any requested id could not be found
        Rad69ServerError
            For any unspecified but valid rad69 error response
        DICOMTrolleyError
            if this is not a valid rad69 error response.

        Notes
        -----
        This just pragmatically interprets the first error returned and assumes the
        rest are the same. Not quite right but let's not overdo it.
        """

        tree = ElementTree.fromstring(response.text)
        errors = tree.findall(RAD69_SOAP_RESPONSE_ERROR_XPATH)
        if not errors:
            raise DICOMTrolleyError(
                f"Could not find any rad69 soap errors in "
                f"response: {response.content[:900]}"
            )

        error_code = errors[0].attrib.get("errorCode")
        error_text = (
            f"Server returns {len(errors)} errors. "
            f"First error: {str(errors[0].attrib)}"
        )
        if error_code == "XDSMissingDocument":
            raise XDSMissingDocumentError(error_text)
        else:
            raise Rad69ServerError(error_text)

    def parse_rad69_response(self, response):
        """Extract datasets out of http response from a rad69 server

        Raises
        ------
        DICOMTrolleyError
            If response is not as expected or if parsing fails

        Returns
        -------
        Iterator[Dataset, None, None]
            All datasets included in this response
        """
        logger.debug("Parsing rad69 response")
        try:
            self.check_for_response_errors(response)
        except DICOMTrolleyError as e:
            self.handle_response_error(e)  # might re-raise
            return None  # error not re-raised. Skip this response # noqa

        part_stream = HTTPMultiPartStream(
            response, stream_chunk_size=self.http_chunk_size
        )
        soap_part = None
        for part in part_stream:
            if not soap_part:
                logger.debug("Discarding initial rad69 soap part")
                soap_part = part  # skip soap part of the response
                continue
            dicom_bytes = part.content
            raw = DicomBytesIO(dicom_bytes)
            try:
                yield dcmread(raw)
            except InvalidDicomError as e:
                raise DICOMTrolleyError(
                    f"Error parsing response as dicom: {e}."
                    f" Response content (first 300 elements) was"
                    f" {str(response.content[:300])}"
                ) from e

    def handle_response_error(self, error):
        """Handle exceptions raised during rad69 request or download"""
        if any(issubclass(type(error), x) for x in self.errors_to_ignore):
            logger.warning(f"Ignoring error on ignore list: {str(error)}")
            return None
        else:
            raise error

    @staticmethod
    def split_instances(instances: Sequence[InstanceReference], num_bins):
        """Split the given instance references into even piles"""
        bin_size = math.ceil(len(instances) / num_bins)
        for i in range(0, len(instances), bin_size):
            yield instances[i : i + bin_size]

    def datasets_async(
        self, instances: Sequence[InstanceReference], max_workers=4
    ):
        """Split instances into chunks and retrieve each chunk in separate thread

        Parameters
        ----------
        instances: Sequence[InstanceReference]
            Retrieve dataset for each of these instances
        max_workers: int, optional
            Use this number of workers in ThreadPoolExecutor. Defaults to
            default for ThreadPoolExecutor

        Notes
        -----
        rad69 allows any number of slices to be combined in one request. The response
        is a chunked multi-part http response with all image data. Requesting each
        slice individually is inefficient. Requesting all slices in one thread might
        limit speed. Somewhere in the middle seems the best bet for optimal speed.
        This function splits all instances between the available workers and lets
        workers process the response streams.

        Raises
        ------
        DICOMTrolleyError
            When a server response cannot be parsed as DICOM

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        # max_workers=None means let the executor figure it out. But for rad69 we
        # still need to determine how many instances to retrieve at once with each
        # worker. Unlimited workers make no sense here. Just use a single thread.
        if max_workers is None:
            max_workers = 1

        with FuturesSession(
            session=self.session,
            executor=ThreadPoolExecutor(max_workers=max_workers),
        ) as futures_session:
            futures = []
            for instance_bin in self.split_instances(instances, max_workers):
                futures.append(
                    futures_session.post(
                        url=self.url,
                        headers=self.post_headers,
                        data=self.create_instances_request(instance_bin),
                    )
                )

            for future in as_completed(futures):
                yield from self.parse_rad69_response(future.result())


class Rad69ServerError(DICOMTrolleyError):
    """Represents a valid error response from a rad69 server"""

    pass


class XDSMissingDocumentError(Rad69ServerError):
    """Some requested ID could not be found on the server"""

    pass



class DICOMTrolleyError(Exception):
    """Base for all exceptions raised in dicomtrolley"""

    pass


class UnSupportedParameterError(DICOMTrolleyError):
    """A query parameter was set that cannot be used in this Searcher type.
    See Query class notes
    """

    pass
